@@ -4,7 +4,7 @@ import numpy as np
 import numba
 import h5py
 
-from exetera.core.readerwriter import DataWriter
+from exetera.core.data_writer import DataWriter
 from exetera.core import utils
 
 
@@ -52,6 +52,13 @@ class Field:
     def chunksize(self):
         return self._field.attrs['chunksize']
 
+    def __bool__(self):
+        # this method is required to prevent __len__ being called on derived methods when fields are queried as
+        #   if f:
+        # rather than
+        #   if f is not None:
+        return True
+
 
 class ReadOnlyFieldArray:
     def __init__(self, field, dataset_name):
@@ -132,7 +139,9 @@ class ReadOnlyIndexedFieldArray:
         self._values_dataset = field[values_name]
 
     def __len__(self):
-        return len(self._index_dataset)-1
+        # TODO: this occurs because of the initialized state of an indexed string. It would be better for the
+        # index to be initialised as [0]
+        return max(len(self._index_dataset)-1, 0)
 
     def __getitem__(self, item):
         try:
@@ -150,6 +159,14 @@ class ReadOnlyIndexedFieldArray:
                         bytestr[index[ir]-np.int64(startindex):
                                 index[ir+1]-np.int64(startindex)].tobytes().decode()
                 return results
+            elif isinstance(item, int):
+                if item >= len(self._index_dataset) - 1:
+                    raise ValueError("index is out of range")
+                start, stop = self._index_dataset[item:item + 2]
+                if start == stop:
+                    return ''
+                value = self._values_dataset[start:stop].tobytes().decode()
+                return value
         except Exception as e:
             print("{}: unexpected exception {}".format(self._field.name, e))
             raise
@@ -210,6 +227,14 @@ class WriteableIndexedFieldArray:
                     rstr = rbytes.decode()
                     results[ir] = rstr
                 return results
+            elif isinstance(item, int):
+                if item >= len(self._index_dataset) - 1:
+                    raise ValueError("index is out of range")
+                start, stop = self._index_dataset[item:item + 2]
+                if start == stop:
+                    return ''
+                value = self._values_dataset[start:stop].tobytes().decode()
+                return value
         except Exception as e:
             print("{}: unexpected exception {}".format(self._field.name, e))
             raise
@@ -355,6 +380,9 @@ class IndexedStringField(Field):
             self._value_wrapper = wrapper(self._field, 'values')
         return self._value_wrapper
 
+    def __len__(self):
+        return len(self.data)
+
 
 class FixedStringField(Field):
     def __init__(self, session, group, name=None, write_enabled=False):
@@ -378,6 +406,9 @@ class FixedStringField(Field):
                 self._value_wrapper = ReadOnlyFieldArray(self._field, 'values')
         return self._value_wrapper
 
+    def __len__(self):
+        return len(self.data)
+
 
 class NumericField(Field):
     def __init__(self, session, group, name=None, write_enabled=False):
@@ -401,6 +432,9 @@ class NumericField(Field):
                 self._value_wrapper = ReadOnlyFieldArray(self._field, 'values')
         return self._value_wrapper
 
+    def __len__(self):
+        return len(self.data)
+
 
 class CategoricalField(Field):
     def __init__(self, session, group,
@@ -427,6 +461,8 @@ class CategoricalField(Field):
                 self._value_wrapper = ReadOnlyFieldArray(self._field, 'values')
         return self._value_wrapper
 
+    def __len__(self):
+        return len(self.data)
 
     # Note: key is presented as value: str, even though the dictionary must be presented
     # as str: value
@@ -459,6 +495,9 @@ class TimestampField(Field):
                 self._value_wrapper = ReadOnlyFieldArray(self._field, 'values')
         return self._value_wrapper
 
+    def __len__(self):
+        return len(self.data)
+
 
 class IndexedStringImporter:
     def __init__(self, session, group, name, timestamp=None, chunksize=None):
@@ -626,7 +665,7 @@ class DateTimeImporter:
                  optional=False, write_days=False, timestamp=None, chunksize=None):
         chunksize = session.chunksize if chunksize is None else chunksize
         timestamp_field_constructor(session, group, name, timestamp, chunksize)
-        self._field = TimestampField(session, group, name, timestamp, write_enabled=True)
+        self._field = TimestampField(session, group, name, write_enabled=True)
         self._results = np.zeros(chunksize , dtype='float64')
         self._optional = optional
 
@@ -671,7 +710,7 @@ class DateImporter:
     def __init__(self, session, group, name,
                  optional=False, timestamp=None, chunksize=None):
         timestamp_field_constructor(session, group, name, timestamp, chunksize)
-        self._field = TimestampField(session, group, name, timestamp, write_enabled=True)
+        self._field = TimestampField(session, group, name, write_enabled=True)
         self._results = np.zeros(chunksize, dtype='float64')
 
         if optional is True:
