@@ -2,6 +2,7 @@ import os
 import uuid
 from datetime import datetime, timezone
 import time
+import warnings
 import numpy as np
 import pandas as pd
 
@@ -39,10 +40,7 @@ from exetera.core import utils
        values are valid or whether they are a standard filtering for different
        categories (age sub-ranges, for example)
  * reader/writers
-   * should be able to read/write through the same object (they can now but not like this)
-     rw = s.get(name)
-     rw.writeable.data[:] = data
-     data = rw.data[:]
+
    * soft sorting / filtering
      rw = s.get(name)
      rw.add_transform(filter)
@@ -50,6 +48,10 @@ from exetera.core import utils
      rw.writeable[:] = data # can this be done?
      rw.clean()
      rw.write(data)
+
+    * refactor: encapsulate dataset handling into session
+      * provide group objects with api
+      * provide field objects with additional api
 """
 
 class Session:
@@ -61,6 +63,59 @@ class Session:
             raise ValueError(error_str.format(type(timestamp)))
         self.chunksize = chunksize
         self.timestamp = timestamp
+        self.datasets = dict()
+
+
+    def __enter__(self):
+        return self
+
+
+    def __exit__(self, etype, evalue, etraceback):
+        self.close()
+
+
+    def open_dataset(self, dataset_path, mode, name):
+        """
+        Open a dataset with the given access mode
+        :param dataset_path: the path to the dataset
+        :param mode: the mode in which the dataset should be opened. This is one of "r", "r+" or "w".
+        :return: The top-level dataset object
+        """
+        h5py_modes = {"r": "r", "r+": "r+", "w": "w"}
+        if name in self.datasets:
+            raise ValueError("A dataset with name '{}' is already open, and must be closed first.".format(name))
+
+        self.datasets[name] = h5py.File(dataset_path, h5py_modes[mode])
+        return self.datasets[name]
+
+
+    def close_dataset(self, name):
+        """
+        Close the dataset with the given name. If there is no dataset with that name, do nothing.
+        :param name: The name of the dataset to be closed
+        :return: None
+        """
+        if name in self.datasets:
+            self.datasets[name].close()
+            del self.datasets[name]
+
+
+    def list_datasets(self):
+        return tuple(n for n in self.datasets.keys())
+
+
+    def get_dataset(self, name):
+        return self.datasets[name]
+
+
+    def close(self):
+        """
+        Close all open datasets
+        :return: None
+        """
+        for v in self.datasets.values():
+            v.close()
+        self.datasets = dict()
 
 
     def get_shared_index(self, keys):
@@ -131,20 +186,30 @@ class Session:
         """
         # TODO: fields is being ignored at present
 
-        readers = tuple(self.get_reader(src_group[f]) for f in keys)
+        readers = tuple(self.get(src_group[f]) for f in keys)
         t1 = time.time()
         sorted_index = self.dataset_sort_index(
-            np.arange(len(readers[0]), dtype=np.uint32), readers)
+            readers, np.arange(len(readers[0].data), dtype=np.uint32))
         print(f'sorted {keys} index in {time.time() - t1}s')
 
         t0 = time.time()
         for k in src_group.keys():
             t1 = time.time()
-            r = self.get_reader(src_group[k])
-            w = r.get_writer(dest_group, k, timestamp, write_mode=write_mode)
-            self.apply_index(sorted_index, r, w)
-            del r
-            del w
+            if src_group != dest_group:
+                r = self.get(src_group[k])
+                w = r.create_like(dest_group, k, timestamp)
+                self.apply_index(sorted_index, r, w)
+                del r
+                del w
+            else:
+                r = self.get(src_group[k]).writeable()
+                if isinstance(r, fld.IndexedStringField):
+                    i, v = self.apply_index(sorted_index, r)
+                    r.indices[:] = i
+                    r.values[:] = v
+                else:
+                    r.data[:] = self.apply_index(sorted_index, r)
+                del r
             print(f"  '{k}' reordered in {time.time() - t1}s")
         print(f"fields reordered in {time.time() - t0}s")
 
@@ -202,7 +267,7 @@ class Session:
             dest_indices, dest_values =\
                 ops.apply_filter_to_index_values(filter_to_apply_,
                                                  src_.indices[:], src_.values[:])
-            if writer_:
+            if writer_ is not None:
                 writer_.indices.write(dest_indices)
                 writer_.values.write(dest_values)
             return dest_indices, dest_values
@@ -234,7 +299,7 @@ class Session:
             dest_indices, dest_values =\
                 ops.apply_indices_to_index_values(index_to_apply_,
                                                   src_.indices[:], src_.values[:])
-            if writer_:
+            if writer_ is not None:
                 writer_.indices.write(dest_indices)
                 writer_.values.write(dest_values)
             return dest_indices, dest_values
@@ -289,61 +354,45 @@ class Session:
         raw_field = None
         raw_fields = None
         if field is not None:
-            val._check_is_reader_or_ndarray('field', field)
-            raw_field = field[:] if isinstance(field, rw.Reader) else field
-        else:
-            raw_fields = []
-            for f in fields:
-                val._check_is_reader_or_ndarray('elements of tuple/list fields', f)
-                raw_fields.append(f[:] if isinstance(f, rw.Reader) else f)
+            raw_field = val.array_from_parameter(self, 'field', field)
+
+        raw_fields = []
+        if fields is not None:
+            for i_f, f in enumerate(fields):
+                raw_fields.append(val.array_from_parameter(self, "'fields[{}]'".format(i_f), f))
         return per._get_spans(raw_field, raw_fields)
 
+
     def _apply_spans_no_src(self, predicate, spans, dest=None):
-        if dest:
-            if val.is_field_parameter(self, 'dest', dest):
-                dest_ = val.field_from_parameter(self, 'dest', dest)
-                results = np.zeros(len(spans) - 1, dtype=dest_.dtype)
-                predicate(spans, results)
-                dest_.data.write(results)
-                return results
-            else:
-                dest_ = val.array_from_parameter(self, 'dest', dest)
-                if len(dest_) != len(spans) - 1:
-                    error_msg = ("if 'dest' (length {}) is an ndarray, it must be one element "
-                                 "shorter than the length of 'spans' (length{})")
-                    raise ValueError(error_msg.format(len(dest_), len(spans)))
-                    predicate(spans, dest_)
-                return dest_
+        assert(dest is None or isinstance(dest, fld.Field))
+        if dest is not None:
+            dest_f = val.field_from_parameter(self, 'dest', dest)
+            results = np.zeros(len(spans) - 1, dtype=dest_f.data.dtype)
+            predicate(spans, results)
+            dest_f.data.write(results)
+            return results
         else:
             results = np.zeros(len(spans) - 1, dtype='int64')
             predicate(spans, results)
             return results
 
     def _apply_spans_src(self, predicate, spans, src, dest=None):
+        assert(dest is None or isinstance(dest, fld.Field))
         src_ = val.array_from_parameter(self, 'src', src)
         if len(src) != spans[-1]:
             error_msg = ("'src' (length {}) must be one element shorter than 'spans' "
                          "(length {})")
             raise ValueError(error_msg.format(len(src_), len(spans)))
 
-        if dest:
-            if val.is_field_parameter(self, 'dest', dest):
-                dest_ = val.field_from_parameter(self, 'dest', dest)
-                results = np.zeros(len(spans) - 1, dtype=dest_.dtype)
-                predicate(spans, results)
-                dest_.data.write(results)
-                return results
-            else:
-                dest_ = val.array_from_parameter(self, 'dest', dest)
-                if len(dest_) != len(spans) - 1:
-                    error_msg = ("if 'dest' (length {}) is an ndarray, it must be one element "
-                                 "shorter than the length of 'spans' (length{})")
-                    raise ValueError(error_msg.format(len(dest_), len(spans)))
-                    predicate(spans, src, dest_)
-                return dest_
+        if dest is not None:
+            dest_f = val.field_from_parameter(self, 'dest', dest)
+            results = np.zeros(len(spans) - 1, dtype=dest_f.data.dtype)
+            predicate(spans, src_, results)
+            dest_f.data.write(results)
+            return results
         else:
             results = np.zeros(len(spans) - 1, dtype=src_.dtype)
-            predicate(spans, src, results)
+            predicate(spans, src_, results)
             return results
 
     def apply_spans_index_of_min(self, spans, src, dest=None):
@@ -358,7 +407,7 @@ class Session:
     def apply_spans_index_of_last(self, spans, dest=None):
         return self._apply_spans_no_src(ops.apply_spans_index_of_last, spans, dest)
 
-    def apply_spans_count(self, spans, dest=None):
+    def apply_spans_count(self, spans, src=None, dest=None):
         return self._apply_spans_no_src(ops.apply_spans_count, spans, dest)
 
     def apply_spans_min(self, spans, src, dest=None):
@@ -397,107 +446,74 @@ class Session:
         dest.complete()
 
 
-    def _aggregate_impl(self, predicate, fkey_indices=None, fkey_index_spans=None,
-                        src=None, dest=None, result_dtype=None):
-        if fkey_indices is None and fkey_index_spans is None:
-            raise ValueError("One of 'fkey_indices' or 'fkey_index_spans' must be set")
-        if fkey_indices is not None and fkey_index_spans is not None:
-            raise ValueError("Only one of 'fkey_indices' and 'fkey_index_spans' may be set")
+    def _aggregate_impl(self, predicate, index, src=None, dest=None):
+        index_ = val.raw_array_from_parameter(self, "index", index)
+
+        dest_field = None
         if dest is not None:
-            if not isinstance(dest, (fld.Field, np.ndarray)):
-                raise ValueError("'writer' must be either a Writer or an ndarray instance")
+            dest_field = val.field_from_parameter(self, "dest", dest)
 
-        if fkey_index_spans is None:
-            fkey_index_spans = self.get_spans(field=fkey_indices)
-
-        if isinstance(dest, np.ndarray):
-            if len(dest) != len(fkey_index_spans) - 1:
-                error = "'writer': ndarray must be of length {} but is of length {}"
-                raise ValueError(error.format(len(fkey_index_spans) - 1), len(dest))
-            elif dest.dtype != result_dtype:
-                raise ValueError(f"'writer' dtype must be {result_dtype} but is {dest.dtype}")
-
-        if isinstance(dest, fld.Field) or dest is None:
-            results = np.zeros(len(fkey_index_spans) - 1, dtype=result_dtype)
-        else:
-            results = dest
+        fkey_index_spans = self.get_spans(field=index)
 
         # execute the predicate (note that not every predicate requires a reader)
-        predicate(fkey_index_spans, src, results)
-
-        if isinstance(dest, fld.Field):
-            dest.data.write(results)
+        results = predicate(fkey_index_spans, src, dest_field)
 
         return dest if dest is not None else results
 
 
-    def aggregate_count(self, fkey_indices=None, fkey_index_spans=None,
-                        src=None, dest=None):
-        return per._aggregate_impl(self.apply_spans_count, fkey_indices, fkey_index_spans,
-                                   src, dest, np.uint32)
+    def aggregate_count(self, index=None, src=None, dest=None):
+        return self._aggregate_impl(self.apply_spans_count, index, src, dest)
 
 
-    def aggregate_first(self, fkey_indices=None, fkey_index_spans=None,
-                        src=None, dest=None):
-        return self.aggregate_custom(self.apply_spans_first, fkey_indices, fkey_index_spans,
-                                     src, dest)
+    def aggregate_first(self, index, src=None, dest=None):
+        return self.aggregate_custom(self.apply_spans_first, index, src, dest)
 
 
-    def aggregate_last(self, fkey_indices=None, fkey_index_spans=None,
-                       src=None, dest=None):
-        return self.aggregate_custom(self.apply_spans_last, fkey_indices, fkey_index_spans,
-                                     src, dest)
+    def aggregate_last(self, index, src=None, dest=None):
+        return self.aggregate_custom(self.apply_spans_last, index, src, dest)
 
 
-    def aggregate_min(self, fkey_indices=None, fkey_index_spans=None,
-                      src=None, dest=None):
-        return self.aggregate_custom(self.apply_spans_min, fkey_indices, fkey_index_spans,
-                                     src, dest)
+    def aggregate_min(self, index, src=None, dest=None):
+        return self.aggregate_custom(self.apply_spans_min, index, src, dest)
 
 
-    def aggregate_max(self, fkey_indices=None, fkey_index_spans=None,
-                      src=None, dest=None):
-        return self.aggregate_custom(self.apply_spans_max, fkey_indices, fkey_index_spans,
-                                     src, dest)
+    def aggregate_max(self, index, src=None, dest=None):
+        return self.aggregate_custom(self.apply_spans_max, index, src, dest)
 
 
-    def aggregate_custom(self,
-                         predicate, fkey_indices=None, fkey_index_spans=None,
-                         src=None, dest=None):
+    def aggregate_custom(self, predicate, index, src=None, dest=None):
         if src is None:
-            raise ValueError("'reader' must not be None")
-        if not isinstance(src, (rw.Reader, np.ndarray)):
-            raise ValueError(f"'reader' must be a Reader or an ndarray but is {type(src)}")
-        required_dtype = src.dtype
-        return per._aggregate_impl(predicate, fkey_indices, fkey_index_spans,
-                                   src, dest, required_dtype)
+            raise ValueError("'src' must not be None")
+        val.ensure_valid_field_like("src", src)
+        if dest is not None:
+            val.ensure_valid_field("dest", dest)
+
+        return self._aggregate_impl(predicate, index, src, dest)
 
 
     def join(self,
              destination_pkey, fkey_indices, values_to_join,
              writer=None, fkey_index_spans=None):
-        if fkey_indices is not None:
-            if not isinstance(fkey_indices, (rw.Reader, np.ndarray)):
-                raise ValueError(f"'fkey_indices' must be a type of Reader or an ndarray")
-        if values_to_join is not None:
-            if not isinstance(values_to_join, (rw.Reader, np.ndarray)):
-                raise ValueError(f"'values_to_join' must be a type of Reader but is {type(values_to_join)}")
-            if isinstance(values_to_join, rw.IndexedStringReader):
-                raise ValueError(f"Joins on indexed string fields are not supported")
 
-        if isinstance(values_to_join, rw.Reader):
-            raw_values_to_join = values_to_join[:]
-        else:
-            raw_values_to_join = values_to_join
+        if isinstance(destination_pkey, fld.IndexedStringField):
+            raise ValueError("'destination_pkey' must not be an indexed string field")
+        if isinstance(fkey_indices, fld.IndexedStringField):
+            raise ValueError("'fkey_indices' must not be an indexed string field")
+        if isinstance(values_to_join, rw.IndexedStringReader):
+            raise ValueError("Joins on indexed string fields are not supported")
+
+        raw_fkey_indices = val.raw_array_from_parameter(self, "fkey_indices", fkey_indices)
+
+        raw_values_to_join = val.raw_array_from_parameter(self, "values_to_join", values_to_join)
 
         # generate spans for the sorted key indices if not provided
         if fkey_index_spans is None:
-            fkey_index_spans = self.get_spans(field=fkey_indices)
+            fkey_index_spans = self.get_spans(field=raw_fkey_indices)
 
         # select the foreign keys from the start of each span to get an ordered list
         # of unique id indices in the destination space that the results of the predicate
         # execution are mapped to
-        unique_fkey_indices = fkey_indices[:][fkey_index_spans[:-1]]
+        unique_fkey_indices = raw_fkey_indices[fkey_index_spans[:-1]]
 
         # generate a filter to remove invalid foreign key indices (where values in the
         # foreign key don't map to any values in the destination space
@@ -509,13 +525,15 @@ class Session:
         safe_values_to_join = raw_values_to_join[invalid_filter]
 
         # now get the memory that the results will be mapped to
-        destination_space_values = writer.chunk_factory(len(destination_pkey))
+        #destination_space_values = writer.chunk_factory(len(destination_pkey))
+        destination_space_values = np.zeros(len(destination_pkey), dtype=raw_values_to_join.dtype)
+
 
         # finally, map the results from the source space to the destination space
         destination_space_values[safe_unique_fkey_indices] = safe_values_to_join
 
         if writer is not None:
-            writer.write(destination_space_values)
+            writer.data.write(destination_space_values)
         else:
             return destination_space_values
 
@@ -562,6 +580,7 @@ class Session:
 
         writer.write(destination_space_values)
 
+
     def get(self, field):
         if isinstance(field, fld.Field):
             return field
@@ -684,13 +703,14 @@ class Session:
         print('  building patient_id index')
         t0 = time.time()
         target_lookup = dict()
-        for i, v in enumerate(target[:]):
+        target_ = val.raw_array_from_parameter(self, "target", target)
+        for i, v in enumerate(target_):
             target_lookup[v] = i
         print(f'  target lookup built in {time.time() - t0}s')
 
         print('  perform initial index')
         t0 = time.time()
-        foreign_key_elems = foreign_key[:]
+        foreign_key_elems = val.raw_array_from_parameter(self, "foreign_key", foreign_key)
         # foreign_key_index = np.asarray([target_lookup.get(i, -1) for i in foreign_key_elems],
         #                                    dtype=np.int64)
         foreign_key_index = np.zeros(len(foreign_key_elems), dtype=np.int64)
@@ -704,8 +724,11 @@ class Session:
             foreign_key_index[i_k] = index
         print(f'  initial index performed in {time.time() - t0}s')
 
-        if destination:
-            destination.write(foreign_key_index)
+        if destination is not None:
+            if val.is_field_parameter(destination):
+                destination.data.write(foreign_key_index)
+            else:
+                destination[:] = foreign_key_index
         else:
             return foreign_key_index
 
@@ -730,61 +753,6 @@ class Session:
         return uid + '.hdf5'
 
 
-    def map_left(self, left_on, right_on, map_mode='both'):
-        valid_map_modes = ('map_to', 'map_from', 'both')
-
-        if map_mode not in valid_map_modes:
-            raise ValueError("'map_mode' must be one of {}".format(valid_map_modes))
-
-        l_key_raw = val.raw_array_from_parameter(self, 'left_on', left_on)
-        l_index = np.arange(len(l_key_raw), dtype=np.int64)
-        l_df = pd.DataFrame({'l_k': l_key_raw, 'l_index': l_index})
-
-        r_key_raw = val.raw_array_from_parameter(self, 'right_on', right_on)
-        r_index = np.arange(len(r_key_raw), dtype=np.int64)
-        r_df = pd.DataFrame({'r_k': r_key_raw, 'r_index': r_index})
-
-        df = pd.merge(left=l_df, right=r_df, left_on='l_k', right_on='r_k', how='left')
-        l_to_r_map = df['l_index'].to_numpy()
-        l_to_r_filt = np.logical_not(df['l_index'].isnull())
-        r_to_l_map = df['r_index'].to_numpy(dtype=np.int64)
-        r_to_l_filt = np.logical_not(df['r_index'].isnull())
-        return l_to_r_map, l_to_r_filt, r_to_l_map, r_to_l_filt
-
-    def map_right(self, left_on, right_on, map_mode='both'):
-        l_key_raw = val.raw_array_from_parameter(self, 'left_on', left_on)
-        l_index = np.arange(len(l_key_raw), dtype=np.int64)
-        l_df = pd.DataFrame({'l_k': l_key_raw, 'l_index': l_index})
-
-        r_key_raw = val.raw_array_from_parameter(self, 'right_on', right_on)
-        r_index = np.arange(len(r_key_raw), dtype=np.int64)
-        r_df = pd.DataFrame({'r_k': r_key_raw, 'r_index': r_index})
-
-        df = pd.merge(left=r_df, right=l_df, left_on='r_k', right_on='l_k', how='left')
-        l_to_r_map = df['r_index'].to_numpy()
-        l_to_r_filt = np.logical_not(df['r_index'].isnull())
-        r_to_l_map = df['l_index'].to_numpy(dtype=np.int64)
-        r_to_l_filt = np.logical_not(df['l_index'].isnull())
-        return l_to_r_map, l_to_r_filt, r_to_l_map, r_to_l_filt
-
-
-    def merge(self, left_on, right_on, how,
-              left_fields=None, right_fields=None,
-              left_writers=None, right_writers=None):
-        if how == 'left':
-            self.merge_left(left_on, right_on, left_fields, right_fields,
-                            left_writers, right_writers)
-        elif how == 'right':
-            self.merge_right(left_on, right_on, left_fields, right_fields,
-                             left_writers, right_writers)
-        elif how == 'inner':
-            self.merge_inner(left_on, right_on, left_fields, right_fields,
-                             left_writers, right_writers)
-        elif how == 'outer':
-            self.merge_outer(left_on, right_on, left_fields, right_fields,
-                             left_writers, right_writers)
-
-
     def merge_left(self, left_on, right_on,
                    right_fields=tuple(), right_writers=None):
         l_key_raw = val.raw_array_from_parameter(self, 'left_on', left_on)
@@ -802,7 +770,8 @@ class Session:
         right_results = list()
         for irf, rf in enumerate(right_fields):
             rf_raw = val.raw_array_from_parameter(self, 'right_fields[{}]'.format(irf), rf)
-            joined_field = per._safe_map(rf_raw, r_to_l_map, r_to_l_filt)
+            joined_field = ops.safe_map(rf_raw, r_to_l_map, r_to_l_filt)
+            # joined_field = per._safe_map(rf_raw, r_to_l_map, r_to_l_filt)
             if right_writers is None:
                 right_results.append(joined_field)
             else:
@@ -828,7 +797,7 @@ class Session:
         left_results = list()
         for ilf, lf in enumerate(left_fields):
             lf_raw = val.raw_array_from_parameter(self, 'left_fields[{}]'.format(ilf), lf)
-            joined_field = per._safe_map(lf_raw, l_to_r_map, l_to_r_filt)
+            joined_field = ops.safe_map(lf_raw, l_to_r_map, l_to_r_filt)
             if left_writers is None:
                 left_results.append(joined_field)
             else:
@@ -856,7 +825,7 @@ class Session:
         left_results = list()
         for ilf, lf in enumerate(left_fields):
             lf_raw = val.raw_array_from_parameter(self, 'left_fields[{}]'.format(ilf), lf)
-            joined_field = per._safe_map(lf_raw, l_to_i_map, l_to_i_filt)
+            joined_field = ops.safe_map(lf_raw, l_to_i_map, l_to_i_filt)
             if left_writers is None:
                 left_results.append(joined_field)
             else:
@@ -865,7 +834,7 @@ class Session:
         right_results = list()
         for irf, rf in enumerate(right_fields):
             rf_raw = val.raw_array_from_parameter(self, 'right_fields[{}]'.format(irf), rf)
-            joined_field = per._safe_map(rf_raw, r_to_i_map, r_to_i_filt)
+            joined_field = ops.safe_map(rf_raw, r_to_i_map, r_to_i_filt)
             if right_writers is None:
                 right_results.append(joined_field)
             else:
@@ -874,10 +843,6 @@ class Session:
         return left_results, right_results
 
 
-    def merge_outer(self, left_on, left_fields, right_on, right_fields):
-        raise NotImplementedError()
-
-
     def _map_fields(self, field_map, field_sources, field_sinks):
         rtn_sinks = None
         if field_sinks is None:
@@ -914,26 +879,8 @@ class Session:
             snk_ = val.field_from_parameter(self, 'field_sinks', snk)
             ops.ordered_map_valid_stream(src_, map_, snk_)
 
-    def ordered_merge(self, left_on, right_on, how,
-                      left_field_sources=tuple(), left_field_sinks=None,
-                      right_field_sources=tuple(), right_field_sinks=None,
-                      left_unique=False, right_unique=False):
-        if how == 'left':
-            return self.ordered_merge_left(left_on, right_on, left_field_sinks, left_unique, left_field_sources,
-                                           right_unique)
-        elif how == 'right':
-            return self.ordered_merge_left(right_on, left_on, right_field_sinks, right_unique, right_field_sources,
-                                           left_unique)
-        elif how == 'inner':
-            return self.ordered_merge_inner(left_on, right_on,
-                                            left_field_sources, left_field_sinks,
-                                            right_field_sources, right_field_sinks,
-                                            left_unique, right_unique)
-        else:
-            raise ValueError("'how' must be one of 'left', 'right' or 'inner'")
 
-
-    def ordered_merge_left(self, left_on, right_on, left_field_sources=tuple(), left_field_sinks=None,
+    def ordered_merge_left(self, left_on, right_on, right_field_sources=tuple(), left_field_sinks=None,
                            left_to_right_map=None, left_unique=False, right_unique=False):
         """
         Generate the results of a left join apply it to the fields described in the tuple
@@ -958,19 +905,20 @@ class Session:
         :return: If left_field_sinks is not set, a tuple of the output fields is returned
         """
         if left_field_sinks is not None:
-            if len(left_field_sources) != len(left_field_sinks):
+            if len(right_field_sources) != len(left_field_sinks):
                 msg = ("{} and {} should be of the same length but are length {} and {} "
                        "respectively")
-                raise ValueError(msg.format(len(left_field_sources), len(left_field_sinks)))
-        val.all_same_basic_type('left_field_sources', left_field_sources)
+                raise ValueError(msg.format(len(right_field_sources), len(left_field_sinks)))
+        val.all_same_basic_type('left_field_sources', right_field_sources)
         if left_field_sinks and len(left_field_sinks) > 0:
             val.all_same_basic_type('left_field_sinks', left_field_sinks)
 
         streamable = val.is_field_parameter(left_on) and \
                      val.is_field_parameter(right_on) and \
-                     val.is_field_parameter(left_field_sources[0]) and \
+                     val.is_field_parameter(right_field_sources[0]) and \
                      left_field_sinks is not None and \
-                     val.is_field_parameter(left_field_sinks[0])
+                     val.is_field_parameter(left_field_sinks[0]) and \
+                     left_to_right_map is not None
 
         result = None
         has_unmapped = None
@@ -985,35 +933,31 @@ class Session:
                     result = left_to_right_map
                 else:
                     result = np.zeros(len(left_on), dtype=np.int64)
+                    left_data = val.array_from_parameter(self, "left_on", left_on)
+                    right_data = val.array_from_parameter(self, "right_on", right_on)
                     has_unmapped = \
-                        ops.ordered_map_to_right_right_unique(left_on, right_on, result)
-
+                        ops.ordered_map_to_right_right_unique(
+                            left_data, right_data, result)
         else:
             if right_unique == False:
                 raise ValueError("Right key must not have duplicates")
-                # if streamable:
-                #     has_unmapped =\
-                #         ops.ordered_map_to_right_left_unique_streamed(left_on, right_on,
-                #                                                       left_to_right_map)
-                #     result = left_to_right_map.data[:]
-                # else:
-                #     result = np.zeros(len(left_on), dtype=np.int64)
-                #     has_unmapped =\
-                #         ops.ordered_map_to_right_left_unique(left_on, right_on, result)
             else:
                 result = np.zeros(len(left_on), dtype=np.int64)
-                has_unmapped = ops.ordered_map_to_right_both_unique(left_on, right_on, result)
+                left_data = val.array_from_parameter(self, "left_on", left_on)
+                right_data = val.array_from_parameter(self, "right_on", right_on)
+                has_unmapped = ops.ordered_map_to_right_both_unique(
+                    left_data, right_data, result)
 
         if streamable:
-            self._streaming_map_fields(result, left_field_sources, left_field_sinks)
+            self._streaming_map_fields(result, right_field_sources, left_field_sinks)
             return None
         else:
-            rtn_left_sinks = self._map_fields(result, left_field_sources, left_field_sinks)
+            rtn_left_sinks = self._map_fields(result, right_field_sources, left_field_sinks)
             return rtn_left_sinks
 
 
     def ordered_merge_right(self, left_on, right_on,
-                            right_field_sources=tuple(), right_field_sinks=None,
+                            left_field_sources=tuple(), right_field_sinks=None,
                             right_to_left_map=None, left_unique=False, right_unique=False):
         """
         Generate the results of a right join apply it to the fields described in the tuple
@@ -1037,15 +981,14 @@ class Session:
         unique values
         :return: If right_field_sinks is not set, a tuple of the output fields is returned
         """
-        return self.ordered_merge_left(right_on, left_on, right_field_sources, right_field_sinks, right_to_left_map,
-                                       right_unique, left_unique)
+        return self.ordered_merge_left(right_on, left_on, left_field_sources, right_field_sinks,
+                                       right_to_left_map, right_unique, left_unique)
 
 
     def ordered_merge_inner(self, left_on, right_on,
                             left_field_sources=tuple(), left_field_sinks=None,
                             right_field_sources=tuple(), right_field_sinks=None,
-                            left_to_inner_map=None, right_to_inner_map=None,
-                            left_unique=False, right_unique=False,):
+                            left_unique=False, right_unique=False):
 
         if left_field_sinks is not None:
             if len(left_field_sources) != len(left_field_sinks):
@@ -1065,29 +1008,24 @@ class Session:
         if right_field_sinks and len(right_field_sinks) > 0:
             val.all_same_basic_type('right_field_sinks', right_field_sinks)
 
+        left_data = val.array_from_parameter(self, 'left_on', left_on)
+        right_data = val.array_from_parameter(self, 'right_on', right_on)
+
         result = None
+        inner_length = ops.ordered_inner_map_result_size(left_data, right_data)
+
+        left_to_inner = np.zeros(inner_length, dtype=np.int64)
+        right_to_inner = np.zeros(inner_length, dtype=np.int64)
         if left_unique is False:
             if right_unique is False:
-                inner_length = ops.ordered_inner_map_result_size(left_on, right_on)
-                left_to_inner = np.zeros(inner_length, dtype=np.int64)
-                right_to_inner = np.zeros(inner_length, dtype=np.int64)
-                ops.ordered_inner_map(left_on, right_on, left_to_inner, right_to_inner)
+                ops.ordered_inner_map(left_data, right_data, left_to_inner, right_to_inner)
             else:
-                inner_length = ops.ordered_inner_map_result_size(left_on, right_on)
-                left_to_inner = np.zeros(inner_length, dtype=np.int64)
-                right_to_inner = np.zeros(inner_length, dtype=np.int64)
-                ops.ordered_inner_map_left_unique(right_on, left_on, right_to_inner, left_to_inner)
+                ops.ordered_inner_map_left_unique(right_data, left_data, right_to_inner, left_to_inner)
         else:
             if right_unique is False:
-                inner_length = ops.ordered_inner_map_result_size(left_on, right_on)
-                left_to_inner = np.zeros(inner_length, dtype=np.int64)
-                right_to_inner = np.zeros(inner_length, dtype=np.int64)
-                ops.ordered_inner_map_left_unique(left_on, right_on, left_to_inner, right_to_inner)
+                ops.ordered_inner_map_left_unique(left_data, right_data, left_to_inner, right_to_inner)
             else:
-                inner_length = ops.ordered_inner_map_result_size(left_on, right_on)
-                left_to_inner = np.zeros(inner_length, dtype=np.int64)
-                right_to_inner = np.zeros(inner_length, dtype=np.int64)
-                ops.ordered_inner_map_both_unique(left_on, right_on, left_to_inner, right_to_inner)
+                ops.ordered_inner_map_both_unique(left_data, right_data, left_to_inner, right_to_inner)
 
         rtn_left_sinks = self._map_fields(left_to_inner, left_field_sources, left_field_sinks)
         rtn_right_sinks = self._map_fields(right_to_inner, right_field_sources, right_field_sinks)
